"""Demonstrate how to capture execution traces.

Each method demonstrates how to capture execution traces using different methods.
It also demonstrates how to add metrics and manual generations to the traces.
Generations are added automatically when using `opper.call` or when functions created
with `opper.functions.create` or `@fn` are called.
"""

import asyncio
import time

from opperai import AsyncOpper, Opper, trace

opper = Opper()


def trace_with_context_manager():
    """trace using context manager"""

    with opper.traces.start(name="context manager") as span:
        span.save_metric("score", 100.0, "context manager")


def trace_manually():
    """trace manually using `start_span`"""

    span = opper.traces.start_span("manually created span")
    span.save_metric("score", 100.0, "manually created span")

    span.end()


@trace
def trace_with_decorator():
    """trace function using the `@trace` decorator"""

    opper.traces.current_span.save_metric("score", 100.0, "decorator")


def manual_generation():
    """manually add a generation to the current span"""

    with opper.traces.start(
        name="manual generation",
        input="What is the meaning of life?",
    ) as span:
        time.sleep(0.1)
        span.save_generation(
            duration_ms=100,
            input="What is the meaning of life?",
            response="42",
            model="openai/gpt-4o",
            messages=[
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "What is the meaning of life?"},
                {"role": "assistant", "content": "42"},
            ],
        )


@trace
def trace_with_call():
    result, response = opper.call(name="python/sdk/bare-minimum", input="Hello, world!")
    response.span.save_metric("score", 100.0, "call")

    return result


@trace
def synchronous_tracing():
    print("running synchronous tracing")

    trace_with_context_manager()
    trace_manually()
    trace_with_decorator()
    manual_generation()
    trace_with_call()

    opper.traces.current_span.save_metric(
        "total_score", 100.0, "metric on the root span"
    )


synchronous_tracing()

aopper = AsyncOpper()


async def async_trace_with_context_manager():
    """trace using context manager"""

    async with aopper.traces.start(name="async context manager") as span:
        await span.save_metric("score", 100.0, "async context manager")


async def async_trace_manually():
    """trace manually using `start_span`"""

    span = await aopper.traces.start_span("manually created span")
    await span.save_metric("score", 100.0, "manually created span")
    await span.end()


@trace
async def async_trace_with_decorator():
    """trace function using the `@trace` decorator"""

    await aopper.traces.current_span.save_metric("score", 100.0, "decorator")


async def async_manual_generation():
    """manually add a generation to the current span"""

    async with aopper.traces.start(
        name="async manual generation",
        input="What is the meaning of life?",
    ) as span:
        time.sleep(0.1)
        await span.save_generation(
            duration_ms=100,
            input="What is the meaning of life?",
            response="42",
            model="openai/gpt-4o",
            messages=[
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "What is the meaning of life?"},
                {"role": "assistant", "content": "42"},
            ],
        )


@trace
async def async_trace_with_call():
    result, response = await aopper.call(
        name="python/sdk/bare-minimum", input="Hello, world!"
    )
    await response.span.save_metric("score", 100.0, "call")

    return result


@trace
async def asynchronous_tracing():
    print("running asynchronous tracing")

    await asyncio.gather(
        async_trace_with_context_manager(),
        async_trace_manually(),
        async_trace_with_decorator(),
        async_manual_generation(),
        async_trace_with_call(),
    )

    await aopper.traces.current_span.save_metric("total_score", 100.0, "chain")


asyncio.run(asynchronous_tracing())



import asyncio
from typing import List

from opperai import AsyncOpper, Opper, trace
from opperai.types import Example
from pydantic import BaseModel, Field

opper = Opper()


def bare_minimum():
    output, _ = opper.call(
        name="python/sdk/bare-minimum",
        instructions="answer the following question",
        input="what are some uses of 42",
    )
    print(output)


def bare_minimum_with_model():
    output, _ = opper.call(
        name="python/sdk/bare-minimum-with-model",
        instructions="answer the following question",
        input="what are some uses of 42",
        model="mistral/mistral-tiny-eu",
    )
    print(output)


def structured_input_output():
    class Number(BaseModel):
        x: int

    output, _ = opper.call(
        name="python/sdk/structured-input-output",
        instructions="given a list of numbers return the largest",
        input=[Number(x=6), Number(x=7)],
        output_type=Number,
    )
    print(output)


def call_with_examples():
    output, _ = opper.call(
        name="python/sdk/call-with-examples",
        instructions="extract the weekday from a text",
        examples=[
            Example(input="Today is Monday", output="Monday"),
            Example(input="Friday is the best day of the week", output="Friday"),
            Example(
                input="Saturday is the second best day of the week", output="Saturday"
            ),
        ],
        input="Wonder what day it is on Sunday",
    )
    print(output)


def call_with_structured_examples():
    class Number(BaseModel):
        x: int = Field(..., ge=0, description="value of the number")

    output, _ = opper.call(
        name="python/sdk/call-with-structured-examples",
        instructions="given a list of numbers return the largest",
        input_type=List[Number],
        input=[Number(x=1), Number(x=12)],
        output_type=Number,
        examples=[
            Example(input=[Number(x=1)], output=Number(x=1)),
            Example(input=[Number(x=2), Number(x=3)], output=Number(x=3)),
        ],
    )
    print(output)


@trace
def synchronous_call():
    print("running synchronous calls")
    bare_minimum()
    bare_minimum_with_model()
    structured_input_output()
    call_with_structured_examples()
    call_with_examples()


synchronous_call()

aopper = AsyncOpper()


async def async_bare_minimum():
    output, _ = await aopper.call(
        name="python/sdk/async-bare-minimum",
        instructions="answer the following question",
        input="what are some uses of 42",
    )
    print(output)


async def async_bare_minimum_with_model():
    output, _ = await aopper.call(
        name="python/sdk/async-bare-minimum-with-model",
        instructions="answer the following question",
        input="what are some uses of 42",
        model="mistral/mistral-tiny-eu",
    )
    print(output)


async def async_structured_input_output():
    class Number(BaseModel):
        x: int

    output, _ = await aopper.call(
        name="python/sdk/async-structured-input-output",
        instructions="given a list of numbers return the largest",
        input=[Number(x=6), Number(x=7)],
        output_type=Number,
    )
    print(output)


async def async_call_with_examples():
    output, response = await aopper.call(
        name="python/sdk/async-call-with-examples",
        instructions="extract the weekday from a text",
        examples=[
            Example(input="Today is Monday", output="Monday"),
            Example(input="Friday is the best day of the week", output="Friday"),
            Example(
                input="Saturday is the second best day of the week", output="Saturday"
            ),
        ],
        input="Wonder what day it is on Sunday",
    )
    print(output)


async def async_call_with_structured_examples():
    class Number(BaseModel):
        x: int = Field(..., ge=0, description="value of the number")

    output, _ = await aopper.call(
        name="python/sdk/async-call-with-structured-examples",
        instructions="given a list of numbers return the largest",
        input_type=List[Number],
        input=[Number(x=1), Number(x=12)],
        output_type=Number,
        examples=[
            Example(input=[Number(x=1)], output=Number(x=1)),
            Example(input=[Number(x=2), Number(x=3)], output=Number(x=3)),
        ],
    )
    print(output)


@trace
async def async_call():
    print("running asynchronous calls")
    await asyncio.gather(
        async_bare_minimum(),
        async_bare_minimum_with_model(),
        async_structured_input_output(),
        async_call_with_examples(),
        async_call_with_structured_examples(),
    )


asyncio.run(async_call())



import pytest
from opperai.types import validate_uuid_xor_path


def test_id_xor_path():
    @validate_uuid_xor_path
    def test_function(uuid=None, path=None):
        return uuid, path

    assert test_function(uuid=1) == (1, None)

    assert test_function(path="test") == (None, "test")

    with pytest.raises(ValueError):
        test_function(uuid=1, path="test")

    with pytest.raises(ValueError):
        test_function()

    with pytest.raises(ValueError):
        test_function(uuid=1, path="test", other="other")



from typing import Dict, List, Optional, Union
from unittest.mock import MagicMock, patch

import pytest
from jsonschema import validate
from opperai import AsyncClient, Client, fn
from opperai.core.utils import convert_function_call_to_json
from opperai.functions.decorator._schemas import type_to_json_schema
from opperai.types import ImageContent
from pydantic import BaseModel


def test_fn_decorator_on_method(client: Client, vcr_cassette):
    class TestClass(BaseModel):
        data: str

        @fn(client=client)
        def test_method(self, text: str) -> str:
            """say hello"""

    test = TestClass(data="Hello")

    res = test.test_method()
    assert "hello" in res.lower()

    res, _ = test.test_method.call("Hello")
    assert "hello" in res.lower()


@pytest.mark.asyncio(scope="module")
async def test_fn_decorator_on_method_async(aclient: AsyncClient, vcr_cassette):
    class TestClass(BaseModel):
        data: str

        @fn(client=aclient)
        async def test_method(self, text: str) -> str:
            """say hello"""

    test = TestClass(data="Hello")

    res = await test.test_method()
    assert "hello" in res.lower()

    res, _ = await test.test_method.call("Hello")
    assert "hello" in res.lower()


def test_fn_decorator_image(client: Client, vcr_cassette):
    class ImageDescription(BaseModel):
        description: str

    @fn(client=client, model="openai/gpt-4o")
    def describe_image(
        image: ImageContent,
    ) -> ImageDescription:
        """given an image describe what it is"""

    description = describe_image(
        ImageContent.from_path("tests/fixtures/images/fossil.png"),
    )
    print(description)

    assert "fossil" in description.description.lower()


@pytest.mark.asyncio(scope="module")
async def test_fn_decorator_image_async(aclient: AsyncClient, vcr_cassette):
    class Word(BaseModel):
        letters: List[str]

    @fn(client=aclient, model="openai/gpt-4o")
    async def extract_letters(
        image: ImageContent,
    ) -> Word:
        """given an image extract the word it represents"""

    word = await extract_letters(
        ImageContent.from_path("tests/fixtures/images/letters.png"),
    )
    print(word)

    assert [x.lower() for x in word.letters] == ["l", "e", "t", "t", "e", "r"]


def hola_in_word(word: str):
    return "hola" in word.lower()


def test_fn_decorator(client: Client, vcr_cassette):
    @fn(client=client)
    def translate(text: str, target_language: str) -> str:
        """Translate text to a target language."""

    translation = translate("Hello", "es")
    assert hola_in_word(translation)

    translation, response = translate.call("Hello", "es")
    assert hola_in_word(translation)

    response.span.save_metric("metric", 1)


@pytest.mark.asyncio
async def test_fn_decorator_async(aclient: AsyncClient, vcr_cassette):
    @fn(client=aclient)
    async def translate(text: str, target_language: str) -> str:
        """Translate text to a target language."""

    translation = await translate("Hello", "es")
    assert hola_in_word(translation)

    translation, response = await translate.call("Hello", "es")
    assert hola_in_word(translation)

    await response.span.save_metric("metric", 1)


@patch("opperai.core._http_clients._http_client.do_request")
def test_decorator(mock_do_request):
    mock_do_request.side_effect = [
        MagicMock(status_code=404),
        MagicMock(
            status_code=200,
            json=lambda: {
                "id": 1,
                "path": "path",
                "description": "description",
                "instructions": "instructions",
            },
        ),
        MagicMock(status_code=200, json=lambda: {"json_payload": "Hola"}),
        MagicMock(status_code=404),
        MagicMock(
            status_code=200,
            json=lambda: {
                "id": 2,
                "path": "path",
                "description": "description",
                "instructions": "instructions",
            },
        ),
        MagicMock(status_code=200, json=lambda: {"json_payload": ["Hola", "Bonjour"]}),
    ]

    @fn()
    def translate(text: str, target_language: str) -> str:
        """Translate text to a target language."""

    @fn()
    def translate_list(text: str, target_languages: List[str]) -> List[str]:
        """Translate text to a list of target languages."""

    assert translate("Hello", "es") == "Hola"
    assert translate_list("Hello", ["es", "fr"]) == ["Hola", "Bonjour"]


@patch("opperai.core._http_clients._http_client.do_request")
def test_decorator_supply_model(mock_do_request):
    mock_do_request.side_effect = [
        MagicMock(status_code=404),
        MagicMock(
            status_code=200,
            json=lambda: {
                "id": 1,
                "path": "path",
                "description": "description",
                "instructions": "instructions",
            },
        ),
        MagicMock(status_code=200, json=lambda: {"json_payload": "Hola"}),
        MagicMock(status_code=404),
        MagicMock(
            status_code=200,
            json=lambda: {
                "id": 2,
                "path": "path",
                "description": "description",
                "instructions": "instructions",
            },
        ),
        MagicMock(status_code=200, json=lambda: {"json_payload": ["Hola", "Bonjour"]}),
    ]

    @fn(model="gpt-4-0125-preview")
    def translate(text: str, target_language: str) -> str:
        """Translate text to a target language."""

    @fn()
    def translate_list(text: str, target_languages: List[str]) -> List[str]:
        """Translate text to a list of target languages."""

    assert translate("Hello", "es") == "Hola"
    assert translate_list("Hello", ["es", "fr"]) == ["Hola", "Bonjour"]


class NestedModel(BaseModel):
    nested_integer: int
    nested_string: str


class ExampleModel(BaseModel):
    integer_field: int
    float_field: float
    string_field: str
    boolean_field: bool
    optional_field: Optional[int]
    list_field: List[int]
    union_field: Union[int, str]
    dict_field: Dict[str, int]
    nested_model: NestedModel
    list_of_models: List[NestedModel]


def test_convert_func_to_json():
    def something(text: str, target_language: str) -> str:
        """Translate text to a target language."""

    input, media = convert_function_call_to_json(something, "Hello", "es")
    assert input == {"text": "Hello", "target_language": "es"}
    assert media == []

    model = ExampleModel(
        integer_field=1,
        float_field=1.0,
        string_field="string",
        boolean_field=True,
        optional_field=1,
        list_field=[1, 2, 3],
        union_field="string",
        dict_field={"key": 1},
        nested_model=NestedModel(nested_integer=1, nested_string="string"),
        list_of_models=[
            NestedModel(nested_integer=1, nested_string="string"),
            NestedModel(nested_integer=1, nested_string="string"),
        ],
    )

    def advanced(text: str, target_language: str, model: ExampleModel) -> str:
        pass

    input, media = convert_function_call_to_json(advanced, "Hello", "es", model)
    assert input == {
        "text": "Hello",
        "target_language": "es",
        "model": model.model_dump(),
    }
    assert media == []


def test_convert_func_to_json_with_image():
    def f(image: ImageContent, text: str) -> str:
        pass

    input, media = convert_function_call_to_json(
        f, ImageContent.from_path("tests/fixtures/images/fossil.png"), "Hello"
    )
    assert input == {"text": "Hello"}
    assert media == [ImageContent.from_path("tests/fixtures/images/fossil.png")]


class ToyModel(BaseModel):
    name: str


class ChildModel(BaseModel):
    toy: ToyModel
    toys: List[ToyModel]


class ParentModel(BaseModel):
    child: ChildModel
    children: List[ChildModel]


@pytest.mark.parametrize(
    "type_input,  example_input",
    [
        (int, 1),
        (str, "string"),
        (float, 1.0),
        (bool, True),
        # (Dict[str, int], {"key": 1}),
        (List[int], [1, 2, 3]),
        (
            ParentModel,
            {
                "child": {
                    "toy": {"name": "name"},
                    "toys": [{"name": "name"}],
                },
                "children": [{"toy": {"name": "name"}, "toys": [{"name": "name"}]}],
            },
        ),
        (
            List[ParentModel],
            [
                {
                    "child": {
                        "toy": {"name": "name"},
                        "toys": [{"name": "name"}],
                    },
                    "children": [{"toy": {"name": "name"}, "toys": [{"name": "name"}]}],
                }
            ],
        ),
        (
            List[List[ParentModel]],
            [
                [
                    {
                        "child": {
                            "toy": {"name": "name"},
                            "toys": [{"name": "name"}],
                        },
                        "children": [
                            {"toy": {"name": "name"}, "toys": [{"name": "name"}]}
                        ],
                    }
                ],
            ],
        ),
    ],
)
def test_type_to_json_schema(type_input, example_input):
    schema = type_to_json_schema(type_input)
    print(schema)
    assert validate(example_input, schema) is None



from unittest.mock import patch

import httpx
import pytest
from opperai import AsyncOpper, Opper
from opperai._client import AsyncClient, Client
from opperai.types import Message
from opperai.types.exceptions import OpperTimeoutError


@pytest.mark.asyncio(scope="module")
async def test_async_client_raises_timeout_error(vcr_cassette):
    async def mock_request(method, path, **kwargs):
        raise httpx.TimeoutException("request timed out")

    opper = AsyncOpper(client=AsyncClient())
    f = await opper.functions.create("test_async_client_raises_timeout_error", "test")

    f._client.http_client.session.request = mock_request

    with pytest.raises(OpperTimeoutError):
        await f.chat(
            messages=[Message(role="user", content="say hello")],
        )

    with patch("opperai.core._http_clients.aconnect_sse") as mock_aconnect_sse:
        mock_aconnect_sse.side_effect = httpx.TimeoutException("request timed out")
        with pytest.raises(OpperTimeoutError):
            r = await f.chat(
                messages=[Message(role="user", content="say hello")],
                stream=True,
            )
            async for _ in r.deltas:
                pass


def test_client_raises_timeout_error(client: Client, vcr_cassette):
    def mock_request(method, path, **kwargs):
        raise httpx.TimeoutException("request timed out")

    opper = Opper(client=client)
    f = opper.functions.create("test_client_raises_timeout_error", "test")

    f._client.http_client.session.request = mock_request

    with pytest.raises(OpperTimeoutError):
        f.chat(messages=[Message(role="user", content="say hello")])

    with patch("opperai.core._http_clients.connect_sse") as mock_connect_sse:
        mock_connect_sse.side_effect = httpx.TimeoutException("request timed out")

        with pytest.raises(OpperTimeoutError):
            r = f.chat(
                messages=[Message(role="user", content="say hello")], stream=True
            )
            for _ in r:
                pass



# ruff: noqa: F401

import base64
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, FilePath, computed_field

from .datasets import DatasetEntry
from .indexes import Document, DocumentIn, Filter, RetrievalResponse
from .spans import SpanMetric
from .validators import validate_uuid_xor_path


class TextMessageContent(BaseModel):
    type: str = "text"
    text: str


class ImageMessageUrl(BaseModel):
    url: str


class ImageMessageFile(BaseModel):
    path: FilePath = Field(exclude=True)

    @computed_field
    @property
    def url(self) -> str:
        path = self.path.as_posix()
        if path.endswith(".png"):
            with open(path, "rb") as image_file:
                base64_image = base64.b64encode(image_file.read()).decode("utf-8")
            return f"data:image/png;base64,{base64_image}"
        elif path.endswith(".jpg"):
            with open(path, "rb") as image_file:
                base64_image = base64.b64encode(image_file.read()).decode("utf-8")
            return f"data:image/jpeg;base64,{base64_image}"
        elif path.endswith(".jpeg"):
            with open(path, "rb") as image_file:
                base64_image = base64.b64encode(image_file.read()).decode("utf-8")
            return f"data:image/jpeg;base64,{base64_image}"
        else:
            raise ValueError(
                "File type not supported. Supported types: .png, .jpg, .jpeg"
            )


class ImageMessageContent(BaseModel):
    type: str = "image_url"
    image_url: Union[ImageMessageUrl, ImageMessageFile]


class MessageContent(BaseModel):
    @classmethod
    def text(cls, text: str) -> TextMessageContent:
        return TextMessageContent(text=text)

    @classmethod
    def image(cls, path: str) -> ImageMessageContent:
        return ImageMessageContent(image_url=ImageMessageFile(path=path))

    @classmethod
    def image_url(cls, url: str) -> ImageMessageContent:
        return ImageMessageContent(image_url=ImageMessageUrl(url=url))


class ImageContent:
    @classmethod
    def from_path(cls, path: str) -> ImageMessageContent:
        return ImageMessageContent(image_url=ImageMessageFile(path=path))

    @classmethod
    def from_url(cls, url: str) -> ImageMessageContent:
        return ImageMessageContent(image_url=ImageMessageUrl(url=url))


class Message(BaseModel):
    role: str
    content: Union[str, List[Union[TextMessageContent, ImageMessageContent]]]


class ChatPayload(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")
    parent_span_uuid: Optional[str] = None

    messages: List[Message]


class FileMetadata(BaseModel):
    file_name: str


class ContextData(BaseModel):
    index_id: int
    content: str
    metadata: Union[Optional[dict], FileMetadata] = Field(default=None)


class StreamingChunk(BaseModel):
    span_id: Optional[str] = None
    delta: Optional[str] = None
    context: Optional[List[ContextData]] = None


class FunctionResponse(BaseModel):
    span_id: Optional[str] = None
    message: Optional[str] = None
    json_payload: Optional[Union[dict, List, Any]] = None
    context: Optional[List[ContextData]] = None
    cached: Optional[bool] = None


class CacheConfiguration(BaseModel):
    exact_match_cache_ttl: Optional[int] = None
    semantic_cache_threshold: Optional[float] = None
    semantic_cache_ttl: Optional[int] = None


class Function(BaseModel):
    uuid: Optional[str] = None
    path: str = Field(
        ...,
        pattern=r"^[a-zA-Z0-9_]+(\/[a-zA-Z0-9_-]+)*$",
        description="Path should not contain characters that could break URLs.",
    )
    description: Optional[str] = None
    input_schema: Optional[Dict[str, Any]] = None
    out_schema: Optional[Dict[str, Any]] = None
    instructions: str
    model: Optional[str] = None
    index_ids: Optional[List[int]] = []
    use_semantic_search: Optional[bool] = None
    few_shot: Optional[bool] = None
    few_shot_count: Optional[int] = None
    cache_configuration: Optional[CacheConfiguration] = None
    dataset_uuid: Optional[str] = None


class Example(BaseModel):
    input: Any
    output: Any


class Error(BaseModel):
    type: str
    message: str
    detail: Any


class Errors(BaseModel):
    errors: List[Error]



from typing import List

from opperai import Index, Opper, trace
from opperai.types import DocumentIn
from opperai.types.indexes import DocumentIn, RetrievalResponse
from pydantic import BaseModel


class Answer(BaseModel):
    steps: List[str]


class QuestionAndContext(BaseModel):
    question: str
    context: List[RetrievalResponse]


@trace
def answer_question(index: Index, question: str) -> Answer:
    results = index.query(question, 1, None)

    result, response = opper.call(
        name="answer_question",
        instructions="Answer the question and provide the steps to do so",
        input=QuestionAndContext(question=question, context=results),
        output_type=Answer,
    )
    response.span.save_metric("artificial_score", 5)

    return result


@trace
def translate(answer: Answer, language: str) -> str:
    result, _ = opper.call(
        name="translate",
        instructions="Translate the answer to the given language",
        input=answer,
        output_type=Answer,
    )
    return result


qna = [
    {
        "question": "I cannot log in to my account",
        "answer": "Use the reset password feature by clicking on 'Forgot password?' and then follow the instructions from email",
        "id": "1",
    },
    {
        "question": "How can I see my invoices?",
        "answer": "Go to the billing section and click on 'Invoices'",
        "id": "2",
    },
    {
        "question": "How can I add a new user to my account?",
        "answer": "Upgrade account and add the user to your account",
        "id": "3",
    },
]

opper = Opper()


def index_qna(qnas: list[dict]):
    index = opper.indexes.create("qna")

    for qna in qnas:
        index.add(
            DocumentIn(
                key=qna["id"],
                content=f"question: {qna['question']}\nanswer: {qna['answer']}",
                metadata={
                    "id": qna["id"],
                },
            )
        )

    return index


def run():
    index = index_qna(qna)

    question = "How can I see my invoices?"

    with opper.traces.start("answer_question") as trace:
        answer = answer_question(index, question)
        print(answer)


run()



from contextlib import asynccontextmanager

import pytest
from httpx import Response
from opperai import AsyncClient
from opperai.types import (
    CacheConfiguration,
    ChatPayload,
    Error,
    Errors,
    Function,
    Message,
    MessageContent,
)
from opperai.types.exceptions import StructuredGenerationError


@asynccontextmanager
async def _function(desc: Function, c: AsyncClient):
    function = await c.functions.create(desc)
    yield function
    await c.functions.delete(uuid=function.uuid)


@pytest.mark.asyncio(scope="module")
async def test_create_function(vcr_cassette, aclient: AsyncClient):
    async with _function(
        Function(
            path="test/sdk/test_create_function_async",
            description="Test function",
            instructions="Do something",
        ),
        aclient,
    ) as function:
        assert function is not None
        assert function.path == "test/sdk/test_create_function_async"


@pytest.mark.asyncio(scope="module")
async def test_get_by_id(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_get_by_id",
            description="Test function",
            instructions="Do something",
        ),
        aclient,
    ) as function:
        f_by_id = await aclient.functions.get(uuid=function.uuid)
        assert f_by_id.path == "test/sdk/test_get_by_id"


@pytest.mark.asyncio(scope="module")
async def test_get_by_path(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_get_by_path",
            description="Test function",
            instructions="Do something",
        ),
        aclient,
    ) as _:
        f_by_path = await aclient.functions.get(path="test/sdk/test_get_by_path")
        assert f_by_path.path == "test/sdk/test_get_by_path"


@pytest.mark.asyncio(scope="module")
async def test_get(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_get",
            description="Test function",
            instructions="Do something",
        ),
        aclient,
    ) as function:
        f_by_id = await aclient.functions.get(uuid=function.uuid)
        assert f_by_id.path == "test/sdk/test_get"

        f_by_path = await aclient.functions.get(path="test/sdk/test_get")
        assert f_by_path.path == "test/sdk/test_get"


@pytest.mark.asyncio(scope="module")
async def test_update_function(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_update_function",
            description="Test function",
            instructions="Do something",
        ),
        aclient,
    ) as function:
        function.instructions = "Do something else"
        updated_function = await aclient.functions.update(function)
        assert updated_function.instructions == "Do something else"


@pytest.mark.asyncio(scope="module")
async def test_delete_function(aclient: AsyncClient, vcr_cassette):
    function = await aclient.functions.create(
        Function(
            path="test/sdk/test_delete_function",
            description="Test function",
            instructions="Do something",
        )
    )
    await aclient.functions.delete(uuid=function.uuid)
    f = await aclient.functions.get(uuid=function.uuid)
    assert f is None


@pytest.mark.asyncio(scope="module")
async def test_image_url(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_image",
            instructions="describe the image",
            model="openai/gpt4-turbo",
        ),
        aclient,
    ) as function:
        f = await aclient.functions.get(uuid=function.uuid)
        resp = await aclient.functions.chat(
            f.path,
            ChatPayload(
                messages=[
                    Message(
                        role="user",
                        content=[
                            MessageContent.image_url(
                                url="https://tf-cmsv2-smithsonianmag-media.s3.amazonaws.com/filer/e1/d3/e1d3fee9-9aaa-4599-ba67-1c71a6d0ed03/1200px-seymouria_fossil.jpg"
                            )
                        ],
                    )
                ]
            ),
        )

        assert "fossil" in resp.message.lower()


@pytest.mark.asyncio(scope="module")
async def test_image_file(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_image",
            instructions="describe the image",
            model="openai/gpt4-turbo",
        ),
        aclient,
    ) as function:
        f = await aclient.functions.get(uuid=function.uuid)
        resp = await aclient.functions.chat(
            f.path,
            ChatPayload(
                messages=[
                    Message(
                        role="user",
                        content=[
                            MessageContent.image(
                                path="tests/fixtures/images/fossil.jpg"
                            )
                        ],
                    )
                ]
            ),
        )

        assert "fossil" in resp.message.lower()


@pytest.mark.asyncio(scope="module")
async def test_async_chat(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_async_chat",
            description="Translate to French",
            instructions="Translate to French",
        ),
        aclient,
    ) as function:
        resp = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )

        assert "bonjour" in resp.message.lower()


@pytest.mark.asyncio(scope="module")
async def test_async_chat_stream(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_async_chat_stream",
            description="Translate to French",
            instructions="Translate to French",
        ),
        aclient,
    ) as function:
        gen = await aclient.functions.chat(
            function.path,
            ChatPayload(messages=[Message(role="user", content="hello")]),
            stream=True,
        )

        resp = "".join([message.delta async for message in gen if message.delta])

        assert "bonjour" in resp.lower()


@pytest.mark.asyncio(scope="module")
async def test_create_function_with_cache(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_create_function_with_cache_async",
            description="Test function",
            instructions="Do something",
            cache_configuration=CacheConfiguration(exact_match_cache_ttl=10),
        ),
        aclient,
    ) as function:
        res = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert not res.cached

        res = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert res.cached


@pytest.mark.asyncio(scope="module")
async def test_create_function_with_cache_flush(aclient: AsyncClient, vcr_cassette):
    async with _function(
        Function(
            path="test/sdk/test_create_function_with_cache_async_flush",
            description="Test function",
            instructions="Do something",
            cache_configuration=CacheConfiguration(exact_match_cache_ttl=10),
        ),
        aclient,
    ) as function:
        res = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert not res.cached

        res = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert res.cached

        await aclient.functions.flush_cache(uuid=function.uuid)
        res = await aclient.functions.chat(
            function.path, ChatPayload(messages=[Message(role="user", content="hello")])
        )
        print(res)
        assert not res.cached


@pytest.mark.asyncio(scope="module")
async def test_failed_structured_generation():
    client = AsyncClient()

    async def mock_request(method, url, **kwargs):
        error = Errors(
            errors=[
                Error(
                    type="StructuredGenerationError",
                    message="test structured generation exception",
                    detail="test structured generation exception",
                )
            ]
        )

        return Response(
            status_code=400,
            json=error.model_dump(),
        )

    client.http_client.session.request = mock_request
    with pytest.raises(StructuredGenerationError):
        await client.functions.chat(
            "fake-path",
            ChatPayload(messages=[Message(role="user", content="hello")]),
        )



import os

from opperai.core.datasets._async_datasets import AsyncDatasets
from opperai.core.datasets._datasets import Datasets
from opperai.core.functions._async_functions import AsyncFunctions
from opperai.core.functions._functions import Functions
from opperai.core.indexes._async_indexes import AsyncIndexes
from opperai.core.indexes._indexes import Indexes
from opperai.core.spans._async_spans import AsyncSpans
from opperai.core.spans._spans import Spans

from .core._http_clients import _async_http_client, _http_client

DEFAULT_API_URL = "https://api.opper.ai"
DEFAULT_TIMEOUT = 120


class AsyncClient:
    functions: AsyncFunctions = None
    indexes: AsyncIndexes = None
    spans: AsyncSpans = None
    datasets: AsyncDatasets = None

    def __init__(
        self,
        api_key: str = None,
        api_url: str = None,
        default_model: str = None,
        timeout: float = DEFAULT_TIMEOUT,
    ):
        if api_key is None:
            api_key = os.getenv("OPPER_API_KEY")
            if api_key is None:
                raise Exception(
                    "API key is not provided and OPPER_API_KEY environment variable is not set."
                )
        if api_url is None:
            api_url = os.getenv("OPPER_API_URL", DEFAULT_API_URL)
        if default_model is None:
            default_model = os.getenv("OPPER_DEFAULT_MODEL")

        self.api_key = api_key
        self.api_url = api_url
        self.default_model = default_model

        self.http_client = _async_http_client(api_key, api_url, timeout)
        self.functions = AsyncFunctions(self.http_client, default_model=default_model)
        self.indexes = AsyncIndexes(self.http_client)
        self.spans = AsyncSpans(self.http_client)
        self.datasets = AsyncDatasets(self.http_client)


class Client:
    functions: Functions
    indexes: Indexes
    spans: Spans
    datasets: Datasets

    def __init__(
        self,
        api_key: str = None,
        api_url: str = None,
        default_model: str = None,
        timeout: float = DEFAULT_TIMEOUT,
    ):
        if api_key is None:
            api_key = os.getenv("OPPER_API_KEY")
            if api_key is None:
                raise Exception(
                    "API key is not provided and OPPER_API_KEY environment variable is not set."
                )
        if api_url is None:
            api_url = os.getenv("OPPER_API_URL", DEFAULT_API_URL)
        if default_model is None:
            default_model = os.getenv("OPPER_DEFAULT_MODEL")

        self.api_key = api_key
        self.api_url = api_url
        self.default_model = default_model

        self.http_client = _http_client(api_key, api_url, timeout)
        self.functions = Functions(self.http_client, default_model=default_model)
        self.indexes = Indexes(self.http_client)
        self.spans = Spans(self.http_client)
        self.datasets = Datasets(self.http_client)
